local myname="ZifnabsTargets"
local Me=CreateFrame("Frame",myname)
if not ZifnabsMod then ZifnabsMod={targets=Me} else ZifnabsMod.targets=Me end
local mod
local L
local db

local MINDISTANCE=0.4				-- only use an additional position slot if distance larger than 0.2 to all existing positions
local MAXIMUMPOSITIONS=10			-- keep max. 10 positions per target in database
local MINTIMEDELTATOKEEP=30*60		-- keep players in db that have been see by the same char at least over a range of 30 minutes
local AGE2DELETEONCE=14*24*60*60	-- delete once met players after 14 days
local PLAYERNAME=UnitName("player").." "..GetRealmName()

local widget={}
local searchresult=nil

-- TODO handling of PVP kills by opponent's faction player

-- --------------------------------------------------------------------------------------
-- INITIALIZE
-- --------------------------------------------------------------------------------------
function Me:Initialize(fwinfo)
	mod=fwinfo.mod
	L=fwinfo.L
	db=fwinfo.db
	if not db.targets then db.targets={list={},settarget={}} end
	
	for name,info in pairs(db.targets.list) do
		if info.player and not info.classified and not info.pvpkills then
			local keep=false
			local seen={}
			for pos,whowhen in pairs(info.seen) do
				for player,time in pairs(whowhen) do
					seen['LAST']=time
					if not seen[player] then seen[player]={first=time,last=time}
					else 
						seen[player].last=time 
						if time-seen[player].first>MINTIMEDELTATOKEEP then keep=true end
					end
				end
			end
			if not keep and time()-seen['LAST']>AGE2DELETEONCE then 
				db.targets.list[name]=nil
				mod.f:PrintR(L.msg.targetdeleted,{name=name})
			end
		end
	end
	widget.frame=mod.ui:newFrame(myname.."_Info",400,20,UIParent,"TOP","TOP",0,0,{transparent=true})
	widget.info=mod.ui:addText(myname.."_Info_Text",400,60,widget.frame,"TOPLEFT","TOPLEFT",4,-4,11,"LEFT","TOP")
	widget.searchbutton=mod.ui:newButton(myname.."_Search_Button",20,16,widget.frame,"TOPLEFT","TOPRIGHT",0,0,{click=Me.ButtonSearch,text="?"})
	widget.searchinput=mod.ui:newEditBox(myname.."_Search_Input",400,20,widget.frame,"TOPLEFT","TOPLEFT",0,0,{onkey=Me.RunSearch,hide=true})
	widget.locatebutton=mod.ui:newButton(myname.."_Locate_Button",20,16,widget.frame,"TOPLEFT","TOPRIGHT",0,-16,{click=Me.ButtonLocate,text="*",hide=true})
	widget.commentbutton=mod.ui:newButton(myname.."_Comment_Button",20,16,widget.frame,"TOPLEFT","TOPRIGHT",-16,0,{click=Me.ButtonComment,text="@",hide=true})
	widget.palbutton=mod.ui:newButton(myname.."_Pal_Button",20,16,widget.frame,"TOPLEFT","TOPRIGHT",-32,0,{click=Me.ButtonPal,text="T",hide=true})
	widget.searchup=mod.ui:newButton(myname.."_Search_Up",20,16,widget.frame,"TOPLEFT","TOPRIGHT",-16,-16,{click=Me.ShowSearch,text="^",hide=true})
	widget.searchdown=mod.ui:newButton(myname.."_Search_Down",20,16,widget.frame,"TOPLEFT","TOPRIGHT",-16,-32,{click=Me.ShowSearch,text="v",hide=true})
	
	mod.core:Register4Event("TICK","targets",Me.TICK)
	mod.core:Register4Event("PLAYER_TARGET_CHANGED","targets",Me.PLAYER_TARGET_CHANGED)

end
function Me:PostInit() 
	Me:SetTarget(db.targets.settarget[PLAYERNAME]) 
	widget.info:SetText(L.msg.notargetyet)
end

-- --------------------------------------------------------------------------------------
-- TICK
-- --------------------------------------------------------------------------------------
function Me:TICK(elapsed)
	if (widget.info.showtime or 0)>0 and not widget.info.fixed then
		widget.info.showtime=widget.info.showtime-elapsed
		if widget.info.showtime<=0 then 
			widget.info:SetText("") 
			widget.locatebutton:Hide()
			widget.commentbutton:Hide()
			widget.palbutton:Hide()
		end
	end
end

-- --------------------------------------------------------------------------------------
-- PLAYER TARGET CHANGED
-- --------------------------------------------------------------------------------------
function Me:PLAYER_TARGET_CHANGED()
	if not widget.info.fixed then widget.info.showtime=5 end
	if UnitExists("target")  then
		local name=UnitName("target")
		local lclass,class=UnitClass("target")
		local player=UnitIsPlayer("target")
		local sex=UnitSex("target")
		local faction=UnitFactionGroup("target")
		local family=UnitCreatureFamily("target") or UnitCreatureType("target")
		local lrace,race=UnitRace("target")
		local level=UnitLevel("target")
		local rarity=UnitClassification("target")
		if rarity=="normal" or rarity=="trivial" or rarity=="minus" then rarity=nil end
		if name~=PLAYERNAME and (not UnitExists("pet") or UnitName("pet")~=name) and (race or not UnitPlayerControlled("target"))then 
			local save={seen={},classified=nil,pvpkills=nil,comment=nil}
			if db.targets.list[name] then save={seen=db.targets.list[name].seen,classified=db.targets.list[name].classified,comment=db.targets.list[name].comment} end
			local x,y,zone,mypos,znr=mod.loc:GetLoc()
			local usepos=nil
			local poscount=0
			local search={distance=999999,oldest=time(),oldestpos=nil,newest=0}
			for pos,whowhen in pairs(save.seen) do
				local pdistance=mod.loc:Distance(mypos,pos)
				poscount=poscount+1
				if pdistance<search.distance then 
					if pdistance<MINDISTANCE then usepos=pos end
					search.distance=pdistance
				end
				local last=0
				for playerchar,timeseen in pairs(whowhen) do 
					last=math.max(last,timeseen)
				end
				search.newest=math.max(search.newest,last)
				if last<search.oldest then
					search.oldest=last
					search.oldestpos=pos
				end
			end
			if not usepos then 
				save.seen[mypos]={}
				usepos=mypos
				if poscount>MAXIMUMPOSITIONS then save.seen[search.oldestpos]=nil end
			end
			save.seen[usepos][PLAYERNAME]=time()
			local guild=GetGuildInfo("target")
			if not player or time()-search.newest>300 or search.distance>1 then
				db.targets.list[name]={
					sex=sex,class=class,player=player,faction=faction,guild=guild,family=family,race=race,lvl=level,rare=rarity,seen=save.seen,classified=save.classified,pvpkills=save.pvpkills,comment=save.comment
				}
			end
			if not widget.info.fixed then Me:ShowInfo(name) end
		end
	end
end

-- --------------------------------------------------------------------------------------
-- GET ENTRY (for external existance check)
-- --------------------------------------------------------------------------------------
function Me:Seen(name) return db.targets.list[name] end

-- --------------------------------------------------------------------------------------
-- SHOW INFO
-- --------------------------------------------------------------------------------------
function Me:ShowSearch(nr,parse) 
	if parse and parse[#parse]=="Up" then
		nr=math.max(1,searchresult.show-1)
	elseif parse and parse[#parse]=="Down" then 
		nr=math.min(searchresult.max,searchresult.show+1) 
	end
	widget.info.fixed=true
	widget.searchbutton:SetText("X")
	if nr>1 then widget.searchup:Show() else widget.searchup:Hide() end
	if nr<searchresult.max then widget.searchdown:Show() else widget.searchdown:Hide() end
	searchresult.show=nr
	Me:ShowInfo(searchresult.name[searchresult.sort[nr]],nr,searchresult.max) 
end
function Me:ShowInfo(name,nr,of)
	local info=db.targets.list[name]
	local met={[PLAYERNAME]={first={time=time()+1},last={time=0}},["others"]={first={time=time()+1},last={time=0}}}
	for pos,whowhen in pairs(info.seen) do
		for playername,timeseen in pairs(whowhen) do
			local realname=playername
			if playername~=PLAYERNAME then playername="others" end
			if timeseen<met[playername].first.time then met[playername].first={time=timeseen,pos=pos,who=realname} end
			if timeseen>met[playername].last.time then met[playername].last={time=timeseen,pos=pos,who=realname} end
		end
	end
	local infotext=" ("..(info.lvl or "??")..")"
	if info.player then 
		infotext=name..", "..(info.guild or L.msg.noguild)..", "..(L.race[info.race] or info.race)..", "..(L.class[info.class] or info.class)..infotext
		if info.classified and info.classified[PLAYERNAME] then infotext=infotext.."\n"..(L.classify[info.classified[PLAYERNAME]] or info.classified[PLAYERNAME]) end
	else 
		infotext=name..", "..(info.race or info.family or "?")..", "..(L.class[info.class] or info.class or "?")..infotext
		if info.rarity then infotext=infotext.." "..info.rarity end
	end
	local usemet=PLAYERNAME
	if not met[PLAYERNAME].first.pos then usemet="others" end
	if met[usemet].first.pos then infotext=infotext.."\n"..Me:Meetinfo(met[usemet].first.time,met[usemet].first.pos,met[usemet].first.who) end
	if met[usemet].last.pos and met[usemet].last.time~=met[usemet].first.time then infotext=infotext.."\n"..Me:Meetinfo(met[usemet].last.time,met[usemet].last.pos,met[usemet].last.who) end
	if info.comment then infotext=infotext.."\n"..info.comment end
	if nr and of then infotext=nr.."/"..of.." "..infotext end
	widget.info:SetText(infotext)
	widget.info.showtime=0
	widget.locatebutton:Show()
	widget.commentbutton:Show()
	if not info.player then widget.palbutton:Hide()
	else	
		widget.palbutton:Show() 
		if not info.classified or not info.classified[PLAYERNAME] then widget.palbutton:SetText("T") else widget.palbutton:SetText(string.sub(info.classified[PLAYERNAME],1,1)) end
	end
	widget.locatebutton.target=name
	widget.commentbutton.target=name
end

function Me:Meetinfo(when,pos,who)
	local x,y,zone,zone_nr=mod.loc:ParseLoc(pos)
	local parsewho=mod.f:Split(who," ")
	if who==PLAYERNAME then who="" else who=" ("..parsewho[1]..")" end
	return date("%d.%m.%y, %H:%M:%S",when).." "..(zone or "nil").." ("..format("%.1f",x).."/"..format("%.1f",y)..")"..who
end

-- --------------------------------------------------------------------------------------
-- UNIT IN MY GUILD
-- --------------------------------------------------------------------------------------
function Me:UnitInMyGuild(who)
	local myguild,myrank,myindex=GetGuildInfo("player")
	local guild,rank,index=GetGuildInfo(who)
	return (guild and myguild==guild)
end

-- --------------------------------------------------------------------------------------
-- CLASSIFY
-- --------------------------------------------------------------------------------------
function Me:Classify(targetname)
	local class=nil
	for p=1,5,1 do
		if UnitExists("party"..p) and UnitName("party"..p)==targetname then class="KNOWN" end
	end
	if db.targets.list[targetname] and db.targets.list[targetname].classified then
		class=db.targets.list[targetname].classified[PLAYERNAME] or class
	end
	return class
end

-- --------------------------------------------------------------------------------------
-- BUTTON SEARCH
-- --------------------------------------------------------------------------------------
function Me:ButtonSearch(button,parse)
	if button:GetText()=="X" then
		widget.searchinput:Hide()
		if widget.commentbutton:GetText()=="S" then
			widget.commentbutton:SetText("@")
		else
			button:SetText("?")
			widget.searchup:Hide()
			widget.searchdown:Hide()
			widget.info.fixed=nil
		end
	else
		widget.searchinput:SetText("")
		widget.searchinput:Show()
		widget.searchinput.request="SEARCH"
	end
end

-- --------------------------------------------------------------------------------------
-- RUN SEARCH
-- --------------------------------------------------------------------------------------
function Me:RunSearch(keywidget,key)
	if key=="ENTER" or key=="ESCAPE" then 
		widget.searchinput:Hide() 
		widget.info.fixed=false
	end
	if key=="ENTER" and widget.searchinput.request=="COMMENT" then Me:StoreComment()
	elseif key=="ENTER" then -- REQUEST SEARCH 
		local search=string.upper(widget.searchinput:GetText())
		local found={}
		local sort={}
		for name,info in pairs(db.targets.list) do
			if (search=="PLAYER" and info.player) or string.find(string.upper(name),search) then 
				local lastseen=0
				for pos,whowhen in pairs(info.seen) do
					for playerchar,timeseen in pairs(whowhen) do lastseen=math.max(lastseen,timeseen) end
				end
				while found[lastseen] do lastseen=lastseen+0.001 end
				sort[1+#sort]=lastseen
				found[lastseen]=name
			end
		end
		if #sort==0 then mod.f:PrintR(L.msg.notfound,{search=search},"buzzer.mp3")
		else
			table.sort(sort)
			searchresult={sort=sort,name=found,max=#sort}
			Me:ShowSearch(#sort)
		end
	end
end

-- --------------------------------------------------------------------------------------
-- BUTTON LOCATE
-- --------------------------------------------------------------------------------------
function Me:ButtonLocate(button,parse) Me:SetTarget(button.target) end
function Me:Click(button,parse) Me:SetTarget(nil) end

function Me:SetTarget(name)	
	db.targets.settarget[PLAYERNAME]=name
	if not name then mod.loc:SetZoneList("targets",nil)
	else
		local myx,myy,myzone,mypos,myznr=mod.loc:GetLoc()
		local tpos=nil
		for pos,whowhen in pairs(db.targets.list[name].seen) do
			local distance=mod.loc:ParseLoc(mypos,pos)
			if not tpos or distance<tpos.distance then
				local tx,ty,tzone,tznr=mod.loc:ParseLoc(pos)
				tpos={distance=distance,zone=tzone,x=tx,y=ty,pos=pos}
			end
		end
		mod.loc:SetZoneList("targets",{[1]={pos=tpos.pos,text=name.."\n"..tpos.zone.." ("..tpos.x.."/"..tpos.y..")",mindist=0,maxdist=9999999,icon=nil,click=Me.Click}})
	end
end

-- --------------------------------------------------------------------------------------
-- BUTTON PAL
-- --------------------------------------------------------------------------------------
function Me:ButtonPal(button,parse)
	local name=widget.commentbutton.target
	local pal=button:GetText()
	if pal=="T" then pal="KNOWN"
	elseif pal=="K" then pal="PAL"
	else pal=nil
	end
	if not db.targets.list[name].classified then db.targets.list[name].classified={} end
	db.targets.list[name].classified[PLAYERNAME]=pal
	button:SetText(string.sub(pal or "T",1,1))
	if searchresult then Me:ShowSearch(searchresult.show) end
end
-- --------------------------------------------------------------------------------------
-- BUTTON COMMENT
-- --------------------------------------------------------------------------------------
function Me:ButtonComment(button,parse)
	if button:GetText()=="S" then
		button:SetText("@")
		widget.info.fixed=nil
		Me:StoreComment()
		Me:ShowSearch(searchresult.show)
	else
		local name=widget.commentbutton.target
		widget.searchinput:SetText(db.targets.list[name].comment or "")
		widget.searchinput:Show()
		widget.searchinput.request="COMMENT"
		button:SetText("S")
		widget.info.fixed=true
	end
end

function Me:StoreComment()
	local comment=widget.searchinput:GetText()
	if comment=="" then comment=nil end
	local name=widget.commentbutton.target
	db.targets.list[name].comment=comment
end